import dataclasses
import numpy

from itertools import groupby
from pathlib import Path
from typing import Dict, List, Optional

import numpy as np
from bs4 import BeautifulSoup

from sumo_docker_pipeline.logger_unit import logger


@dataclasses.dataclass
class MatrixObject(object):
    """
    matrix (2d-array): a 2d-array which contains {value_type}.
    detectors (1d-array): a 1d-array which contains an id of detectors.
    interval_begins (1d-array): a 1d-array which contains a start of an interval-time.
    value_type: name of type that a matrix object has.
    time_interval: length of time-interval.
    """
    matrix: numpy.ndarray
    detectors: numpy.ndarray
    interval_begins: numpy.ndarray
    interval_end: numpy.ndarray
    value_type: str

    def to_npz(self, path_npz: Path):
        dict_obj = dataclasses.asdict(self)
        np.savez(path_npz, **dict_obj)

    @classmethod
    def from_npz(cls, path_npz: Path) -> "MatrixObject":
        data = dict(np.load(str(path_npz), allow_pickle=True))
        return MatrixObject(**data)


@dataclasses.dataclass
class ResultFile(object):
    path_file: Path
    name_file: Optional[str] = None
    tree_object: Optional[BeautifulSoup] = None
    key_attributes: Optional[List[str]] = None

    def __str__(self):
        return self.path_file

    def __post_init__(self):
        self.name_file = self.path_file.name
        self.tree_object = self.load_file()
        self.key_attributes = self.get_attributes(self.tree_object)

    def load_file(self) -> BeautifulSoup:
        f_obj = open(self.path_file, 'r')
        soup = BeautifulSoup(f_obj, 'lxml')
        return soup

    @staticmethod
    def get_attributes(root_soup: BeautifulSoup) -> List[str]:
        __ = []
        for elem in root_soup.find_all('interval'):
            __ += list(elem.attrs.keys())
        # end for
        return list(set(__))

    @staticmethod
    def matrix_with_autofill(matrix_stack: List[List[float]]) -> numpy.ndarray:
        """auto-fill a matrix object with nan value if lengths of lists are different.

        :param matrix_stack: 2nd list. [[value]]
        :return: 2nd ndarray.
        """
        max_length = max([len(l) for l in matrix_stack])
        min_length = min([len(l) for l in matrix_stack])
        if max_length == min_length:
            matrix_value = numpy.array(matrix_stack)
            return matrix_value
        else:
            matrix_value = numpy.zeros([len(matrix_stack), max_length])
            logger.warning('The output file different length of elements. I replaced insufficient values with Nan. '
                           'Be careful the existence of Nan values.')
            matrix_value[:] = np.NAN
            for i, j in enumerate(matrix_stack):
                matrix_value[i][0:len(j)] = j
            # end for
            return matrix_value

    def xml2matrix(self,
                   root_soup: BeautifulSoup,
                   target_element: str) -> MatrixObject:
        """generates matrix object with the specified key name.

        :param root_soup: xml object.
        :param target_element: a name of key which corresponds to values of the matrix.
        :return: MatrixObject
        """
        stacks = []
        __time_interval: Optional[float] = None
        for elem in root_soup.find_all('interval'):
            detector_id = elem.get('id')
            time_begin = float(elem.get('begin'))
            time_end = float(elem.get('end'))

            obj_value = elem.get(target_element)
            try:
                if obj_value == '':
                    target_value = 0.0
                elif obj_value is None:
                    target_value = 0.0
                else:
                    target_value = float(elem.get(target_element))
            except ValueError:
                raise SystemError(f'unexpected error during parsing values because of {obj_value}')
            except KeyError:
                keys = elem.attrs.keys()
                raise KeyError(f'Invalid key name. Available keys are {keys}')
            # end try
            stacks.append([detector_id, time_begin, time_end, target_value])
        # end for
        seq_detector_id = []
        matrix_stack = []
        seq_begin = []
        seq_end = []
        for detector_id, g_obj in groupby(sorted(stacks, key=lambda t: t[0]), key=lambda t: t[0]):
            __ = list(sorted([t for t in g_obj], key=lambda t: t[1]))
            seq_begin = [t[1] for t in __]
            seq_end = [t[2] for t in __]
            seq_value = [t[3] for t in __]
            seq_detector_id.append(detector_id)
            matrix_stack.append(seq_value)
        # end for
        detectors = numpy.array(seq_detector_id)

        begin_time_vector = numpy.array(seq_begin)
        end_time_vector = numpy.array(seq_end)
        matrix_value = self.matrix_with_autofill(matrix_stack)
        assert len(matrix_value.shape) == 2, f'The method expects 2nd array. But it detects {matrix_value.shape} object. ' \
                                             f'Check your xml file at {self.path_file}'
        return MatrixObject(
            matrix=matrix_value,
            detectors=detectors,
            interval_begins=begin_time_vector,
            interval_end=end_time_vector,
            value_type=target_element)

    def to_array_objects(self, aggregation_on: str) -> MatrixObject:
        matrix_obj = self.xml2matrix(root_soup=self.tree_object, target_element=aggregation_on)
        return matrix_obj


@dataclasses.dataclass
class SumoResultObjects(object):
    log_message: str
    path_output_dir: Path
    result_files: Optional[Dict[str, ResultFile]] = None

    def __post_init__(self):
        assert Path(self.path_output_dir).exists()
        xml_files = list(Path(self.path_output_dir).glob('*xml'))
        assert len(xml_files) > 0, f'No output xml found in {self.path_output_dir}'
        self.result_files = {}
        for path_xml in xml_files:
            self.result_files[path_xml.name] = ResultFile(path_xml)
        # end for



import docker
import lxml.etree
import re
from typing import Dict, Any
from pathlib import Path, WindowsPath
from datetime import datetime
from sumo_docker_pipeline.result_module import SumoResultObjects, ResultFile
from sumo_docker_pipeline.logger_unit import logger

from docker.errors import NotFound


class SumoDockerController(object):
    def __init__(self,
                 image_name: str = "kensukemi/sumo-ubuntu18",
                 container_name_base: str = "sumo-docker",
                 mount_dir_host: str = "mount_dir",
                 mount_dir_container: str = "/mount_dir",
                 sumo_command: str = "sumo",
                 device_rerouting_threads: int = 4,
                 is_rewrite_windows_path: bool = True):
        self.image_name = image_name
        self.container_name_base = container_name_base
        self.mount_dir_container = mount_dir_container
        self.mount_dir_host = mount_dir_host
        self.sumo_command = sumo_command
        self.device_rerouting_threads = device_rerouting_threads
        self.is_auto_remove = True

        self.client = docker.from_env()
        self.__check_connection()
        self.is_auto_remove = True
        self.is_rewrite_windows_path = is_rewrite_windows_path

    def __generate_tmp_container_name(self) -> str:
        c_name = f'{self.container_name_base}-{datetime.now().timestamp()}'
        return c_name

    def __check_connection(self):
        c_name = self.__generate_tmp_container_name()
        try:
            command_message = self.client.containers.run(image=self.image_name,
                                                         command=self.sumo_command, name=c_name,
                                                         auto_remove=self.is_auto_remove)
        except docker.errors.NotFound:
            self.is_auto_remove = False
            command_message = self.client.containers.run(image=self.image_name,
                                                         command=self.sumo_command, name=c_name,
                                                         auto_remove=self.is_auto_remove)
        # end except
        assert "German Aerospace Center" in command_message.decode('utf-8')

    def detect_mount_point(self):
        candidate_mount_points = [str(Path(self.mount_dir_container).parent), '/home', '/tmp']
        for path_candidate in candidate_mount_points:
            container_name = self.__generate_tmp_container_name()
            command_message = self.client.containers.run(image=self.image_name,
                                                         command=f'ls {path_candidate}', name=container_name,
                                                         auto_remove=self.is_auto_remove,
                                                         volumes={
                                                             self.mount_dir_host: {'bind': f'{path_candidate}/mount_dir',
                                                                                   'mode': 'rw'}})
            if self.is_auto_remove is False:
                self.delete_container(container_name)
            # end if
            if 'mount_dir' in command_message.decode('utf-8'):
                return path_candidate
            # end if
        # end for
        raise Exception(f'Failed to mount host-directory into container-directory. '
                        f'Try to change mount_dir_container argument of init. Current value is {self.mount_dir_container}')

    def delete_container(self, container_name: str):
        for c in self.client.containers.list(all=True):
            if c.name == container_name:
                try:
                    c.remove()
                except Exception as e:
                    logger.warning(f'We failed to remove container = {container_name}. We recommend to remove it manually. '
                                   f'The reason is {e}')

    def get_sumo_version(self) -> str:
        command_message = self.client.containers.run(image=self.image_name,
                                                     command=f'{self.sumo_command} -V',
                                                     name=self.__generate_tmp_container_name(),
                                                     auto_remove=self.is_auto_remove)
        return command_message.decode('utf-8')

    @staticmethod
    def extract_output_options(config_file_name: Path) -> Dict[str, ResultFile]:
        """List up files in output directory.

        :param config_file_name:
        :return:
        """
        tree = lxml.etree.parse(config_file_name.open())
        root = tree.getroot()
        r_files_type = []
        for t in root.find('output'):
            element_name = t.tag
            value_name: Dict[str, Any] = t.attrib
            if isinstance(value_name['value'], bool):
                continue
            if value_name['value'] in ('true', 'false'):
                continue
            # end if
            try:
                float(value_name['value'])
                continue
            except ValueError:
                if 'prefix' in element_name:
                    output_dir_name = config_file_name.parent.joinpath(''.join(value_name['value'].split('/')[:-1]))
                else:
                    output_dir_name = config_file_name.parent.joinpath((value_name['value']))
                # end if
                for p_obj in output_dir_name.glob('*'):
                    if p_obj.is_dir():
                        continue
                    # enf if
                    r_files_type.append(ResultFile(p_obj))
                # end for
            # end try
        # end for
        return {f.name_file: f for f in r_files_type}

    def check_output_dir(self, target_scenario_name: str, result_file_types: Dict[str, ResultFile]):
        for tag_name, file_type_obj in result_file_types.items():
            dir_name = file_type_obj.path_file
            # end if
            confirm_target = Path(self.mount_dir_host).joinpath(target_scenario_name).joinpath(dir_name)
            if confirm_target.exists() is False:
                logger.info(f'created a directory {confirm_target}')
                confirm_target.mkdir()
            # end if
        # end for

    @staticmethod
    def __extract_output_dir(path_config_file: Path) -> Path:
        with open(path_config_file, 'r') as f:
            tree = lxml.etree.parse(f)
        # end with
        root = tree.getroot()
        output_prefix_element = root.find('output').find('output-prefix')
        assert output_prefix_element is not None, \
            'output-prefix element does not exist in config file. Check your config file.'
        return Path(path_config_file).parent.joinpath(output_prefix_element.attrib['value'])

    @staticmethod
    def rewrite_windows_path(mount_dir_host: str) -> str:
        """rewrite from Windows style into Unix style.
        Ex. C:\\Users\\kensu\\AppData\\Local\\Temp\\tmp3t95r_eb -> /c/Users/kensu/AppData/Local/Temp/tmp3t95r_eb
        """
        unix_form = Path(mount_dir_host).as_posix()
        drive_prefix = re.search(r'^[C-D]\:', str(unix_form))
        if drive_prefix:
            without_prefix = unix_form[drive_prefix.span()[1]:]
            replaced_prefix = '/' + unix_form[drive_prefix.span()[0]:drive_prefix.span()[1]].lower().replace(':', '')
            path_updated = replaced_prefix + without_prefix
        else:
            path_updated = str(unix_form)
        # end if
        return path_updated

    def start_job(self, target_scenario_name: str, config_file_name: str = 'sumo.cfg') -> SumoResultObjects:
        c_name = self.__generate_tmp_container_name()
        path_config_file = Path(self.mount_dir_container).joinpath(target_scenario_name).joinpath(config_file_name)
        if self.is_rewrite_windows_path and isinstance(path_config_file, WindowsPath):
            # If windows...Path structure is broken. Fix it manually.
            path_config_file = path_config_file.as_posix()
        # end if
        job_command = f'{self.sumo_command} -c {path_config_file}'
        if self.device_rerouting_threads > 0:
            job_command += f' --device.rerouting.threads {self.device_rerouting_threads}'
        # end if

        if self.is_rewrite_windows_path and isinstance(Path('./'), WindowsPath):
            # from windows style path into Unix style path. Docker does not accept Windows format.
            logger.info('I replaced a format of source directory in the host side. '
                        'Check it if there is an unknown issue.')
            logger.info(f'Before {self.mount_dir_host}')
            __mount_dir_host = self.rewrite_windows_path(self.mount_dir_host)
            logger.info(f'After {__mount_dir_host}')
        else:
            __mount_dir_host = str(self.mount_dir_host)
        # end if
        logger.debug(f'executing job with command {job_command}')

        command_message = self.client.containers.run(image=self.image_name,
                                                     command=job_command, name=c_name, auto_remove=self.is_auto_remove,
                                                     volumes={__mount_dir_host: {'bind': self.mount_dir_container,
                                                                                 'mode': 'rw'}})
        path_config_file_host = Path(self.mount_dir_host).joinpath(target_scenario_name).joinpath(config_file_name)
        result_file_types = self.extract_output_options(path_config_file_host)
        self.check_output_dir(target_scenario_name, result_file_types)
        res_obj = SumoResultObjects(
            log_message=command_message.decode('utf-8'),
            result_files=result_file_types,
            path_output_dir=self.__extract_output_dir(path_config_file_host)
        )
        if self.is_auto_remove is False:
            self.delete_container(container_name=c_name)
        # end if
        return res_obj

    # ---
    # methods for async methods

    def async_start_job(self):
        raise NotImplementedError()

    def async_check_job_status(self):
        raise NotImplementedError()

    def async_get_job_log(self):
        raise NotImplementedError()
